"""
Freedom dispatch queue

The dispatch queue is used for dispatching messages between threads.

@copyright: 2011 Nir Soffer <nirsof@gmail.com>, Erez Waisbard <waisbard@gmail.com>
@license: GNU GPL, see COPYING for details.
"""

import collections
import threading
import logging

STOP = object()

class queue(object):
    """ Dispatch message to delegate in a background thread 
    
    A delegate may implement this interface:

        def queue_start(self):
            ''' Called when a queue thread is running. Use to perofm
            initialization that must happen in the queue thread. '''
        
        def queue_handle(self, msg):
            ''' Called when a message was dispatched to the queue in the queue
            thread. '''

        def queue_stop(self):
            ''' Called when the queue theard is stopped. Use to perform cleanup
            tasks. '''
    """

    def __init__(self, delegate):
        self._delegate = delegate
        self._queue = collections.deque()
        self._thread = threading.Thread(target=self._loop)
        self._thread.setDaemon(True)
        self._condition = threading.Condition()

    def start(self):
        self._thread.start()

    def stop(self):
        self.handle(STOP)

    def handle(self, msg):
        with self._condition:
            self._queue.append(msg)
            self._condition.notify()

    def _loop(self):
        logging.debug('dispatch.queue running')
        if hasattr(self._delegate, 'queue_start'):
            self._delegate.queue_start()
        while 1:
            with self._condition:
                logging.debug('diapatch.queue waiting for message...')
                while not self._queue:
                    self._condition.wait()
                msg = self._queue.popleft()
            if msg is STOP:
                logging.debug('dispatch.queue received stop signal')
                break
            if hasattr(self._delegate, 'queue_handle'):
                #logging.debug('diapatch.queue handling message %r', msg)
                try:
                    self._delegate.queue_handle(msg)
                except Exception:
                    logging.exception('delegate error')
        if hasattr(self._delegate, 'queue_stop'):
            self._delegate.queue_stop()
        with self._condition:
            self._queue.clear()
        logging.debug('dispatch.queue shutting down')
