""" 
Freedom IMAP4 listener 

Listeners use IMAP IDLE command to wait for messages from the IMAP4 server,
fetch and delete the messages and pass them to the next handler.

@copyright: 2011 Nir Soffer <nirsof@gmail.com>, Erez Waisbard <waisbard@gmail.com>
@license: GNU GPL, see COPYING for details.
"""

import logging
import email
import base64
import time

import imaplib2 as imaplib

from freedom import dispatch
from freedom import event
from freedom import request

class listener(object):
   
    # Time to wait becore reconnecting to the server
    min_reconnect_timeout = 1.0
    max_reconnect_timeout = 120.0

    def __init__(self, host, user, password, next, debug=False):
        self._host = host
        self._user = user
        self._password = password
        self._next = next
        self._debug = debug
        self._queue = dispatch.queue(self)
        self._conn = None
        self._reconnect_timeout = self.min_reconnect_timeout

    ## Running

    def start(self):
        self._queue.start()

    def stop(self):
        self._queue.stop()
        if self._conn:
            # Abort current socket operation - otherwise the queue may stop in
            # 30 mintues.
            try:
                self._conn.socket().close()
            except socket.error:
                logging.warn('error closing socket: %s', e)

    ## dispatch.queue delegate methods

    def queue_start(self):
        logging.info('imap.listener running')
        self._queue.handle('listen')

    def queue_handle(self, msg):
        try:
            if self._conn is None:
                self._connect()
            self._update()
            self._idle()
        except imaplib.IMAP4.abort, e:
            logging.warning('imap.listener connection was aborted: %s', e)
            self._disconnect()
            self._wait_before_reconnect()
        except imaplib.IMAP4.readonly:
            logging.warning('imap.listener mailbox became readonly')
        except Exception:
            logging.exception('imap.listener error')

        # This message must be sent or the listener will stop!
        self._queue.handle(msg)

    def queue_stop(self):
        self._disconnect()
        logging.info('imap.listener shuting down')

    ## Talking with IMAP server

    def _connect(self):
        logging.info('imap.listener connecting to %r', self._host)
        self._conn = imaplib.IMAP4_SSL(self._host) 
        if self._debug:
            self._conn.debug = 4
        logging.info('imap.listener logging in as %r', self._user)
        self._conn.login(self._user, self._password)
        logging.debug('imap.listener selecting INBOX')
        self._conn.select()
        self._reconnect_timeout = self.min_reconnect_timeout
    
    def _disconnect(self):
        if self._conn:
            logging.info('imap.listener disconnecting')
            try:
                self._conn.close()
                self._conn.logout()
            except (imaplib.IMAP4.abort, imaplib.IMAP4.readonly):
                pass
            except Exception:
                logging.exception('error disconnecting')
            self._conn = None

    def _wait_before_reconnect(self):
        logging.info('imap.listener reconnecting in %.0f seconds...', self._reconnect_timeout)
        time.sleep(self._reconnect_timeout)
        # Be polite to the server
        if self._reconnect_timeout < self.max_reconnect_timeout:
            self._reconnect_timeout *= 2

    def _idle(self):
        logging.debug('imap.listener idling')
        # This can block up to 29 minutes
        self._conn.idle()

    ## Subclass required methods

    def _update(self):
        raise NotImplementedError

class request_listener(listener):

    def _update(self):
        logging.debug('imap.request_listener updating')
        while 1:
            uids = self._search_requests()
            if not uids:
                break
            for uid in uids:
                req = self._fetch_request(uid)
                if req:
                    # XXX Drop old requests
                    logging.info('imap.request_listener received request: %r', req)
                    self._next.handle(req)

    def _search_requests(self):
        logging.debug('imap.request_listener searching requests')
        start = time.time()
        # XXX sometime this get stuck
        self._conn.noop()
        t, d = self._conn.uid('SEARCH', '(SUBJECT "http://")')
        uids = d[0].split()
        logging.debug('clock: search_requests: %.2fs', time.time() - start)
        return uids

    def _fetch_request(self, msguid):
        logging.debug('imap.request_listener fetching request: %r', msguid)
        start = time.time()
        t, d = self._conn.uid('FETCH', msguid, '(BODY[HEADER.FIELDS (From Subject)])')
        self._conn.uid('STORE', msguid, '+FLAGS', '\\Deleted')
        result = email.message_from_string(d[0][1])
        try:
            req = request.parse(result['subject'])
        except request.error, e:
            logging.debug('imap.request_listener: %s', e)
            return None
        # XXX Can this fail?
        req['from'] = result['from']
        event.log('server.receive', start)
        logging.debug('clock: fetch_request: %.2fs', time.time() - start)
        return req

class response_listener(listener):

    def _update(self):
        logging.debug('imap.response_listener updating')
        while 1:
            uids = self._search_responses()
            if not uids:
                break
            for uid in uids:
                res = self._fetch_response(uid)
                logging.info('imap.listener received response: %r length: %d',
                        res['url'], len(res['data']))
                self._next.handle(res)

    def _search_responses(self):
        logging.debug('imap.response_listener searching responses')
        start = time.time()
        self._conn.noop()
        t, d = self._conn.uid('SEARCH', '(SUBJECT "Re: http://")')
        uids = d[0].split()
        logging.debug('clock: search_responses: %.2fs', time.time() - start)
        return uids
    
    def _fetch_response(self, uid):
        logging.debug('imap.response_listener fetching response: %r', uid)
        start = time.time()
        event.log('client.receive', start)
        t, d = self._conn.uid('FETCH', uid, 
                '(BODY[TEXT] BODY[HEADER.FIELDS (x-freedom-url x-freedom-filename)])')
        self._conn.uid('STORE', uid, '+FLAGS', '\\Deleted')
        # XXX Seems that BODY[TEXT] is alwways first, but we should
        # verify this using d[0][0].
        header = email.message_from_string(d[1][1])
        body = d[0][1]
        try:
            data = base64.b64decode(body)
        except TypeError, e:
            raise error('Invalid response format: %s' % e)
        res = {}
        res['data'] = data
        res['url'] = header['x-freedom-url']
        res['filename'] = header['x-freedom-filename']
        logging.debug('clock: fetch_response: %.2fs', time.time() - start)
        return res 

