import os
import tarfile
import StringIO
import uuid

from freedom import response

host = 'example.org'
index = os.path.join(host, 'page', 'index.html')
html_data = '''
<html>
<head>
<link rel="stylesheet" href="/css/style.css">
<link rel="stylesheet" href="/css/screen.css" media="Screen, Print">
<link rel="stylesheet" href="/css/all.css" media="all">
<link rel="stylesheet" href="/css/print.css" media="print">
<link rel="stylesheet" href="/css/missing.css">
<script type="text/javascript"></script>
<script type="text/javascript" src="script1.js"></script>
</head>
<body>
<h1>Example page</h1>
<img src="image1.jpg">
<img src="image1.jpg">
<img src="/images/image2.png">
<img src="missing.gif">
<img class="nosrc">
<script type="text/javascript"></script>
<script type="text/javascript" src="script2.js"></script>
</body>
</html>
'''
xhtml_data = '''
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" 
        "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html>
<head>
<link rel="stylesheet" href="/css/style.css" />
<script type="text/javascript"></script>
<script type="text/javascript" src="script1.js"></script>
</head>
<body>
<script type="text/javascript"></script>
<script type="text/javascript" src="script2.js"></script>
</body>
</html>
'''
style = os.path.join(host, 'css', 'style.css')

# This stylesheet must be escaped when embeding it inside XHTML
style_data = '''
p :before   { content: "< " }
p :after    { content: " &" }
'''

style_screen = os.path.join(host, 'css', 'screen.css')
style_screen_data = '#screen {color: red;}'

style_all = os.path.join(host, 'css', 'all.css')
style_all_data = '#all {color: red;}'

style_print = os.path.join(host, 'css', 'print.css')
style_print_data = '#print {color: red;}'

# This image is locatd in the same directory as the html
image1 = os.path.join(host, 'page', 'image1.jpg')
image1_data = 'image 1 data'

# This image is located in an upper directory
image2 = os.path.join(host, 'images', 'image2.png')
image2_data = 'image 2 data'

def create_page(tmpdir, type='html'):
    index_data = globals()[type + '_data']
    tmpdir.join(index).ensure().write(index_data)
    tmpdir.join(style).ensure().write(style_data)
    tmpdir.join(style_screen).ensure().write(style_screen_data)
    tmpdir.join(style_all).ensure().write(style_all_data)
    tmpdir.join(style_print).ensure().write(style_print_data)
    tmpdir.join(image1).ensure().write(image1_data)
    tmpdir.join(image2).ensure().write(image2_data)

class fake_uuid:
    def __init__(self):
        self.count = 0
    def uuid4(self):
        self.count += 1
        return str(self.count)

def setup_function(f):
    response.uuid = fake_uuid()

def teardown_function(f):
    response.uuid = uuid

def test_default_builder(tmpdir):
    create_page(tmpdir)
    req = {'url': 'http://gnu.org'}
    builder = response.builder(req)
    assert builder.__class__ is response.multipart_builder

def test_attachemnt_builder(tmpdir):
    create_page(tmpdir)
    req = {'url': 'http://gnu.org', 'attach': True}
    builder = response.builder(req)
    assert builder.__class__ is response.tar_builder

def test_tar_builder(tmpdir):
    create_page(tmpdir)
    builder = response.tar_builder()
    msg = builder.build(str(tmpdir), index)
    assert not msg.is_multipart()
    assert msg['content-type'] == 'application/octet-stream'
    assert msg['content-disposition'] == 'attachment; filename=page.bin'
    data = msg.get_payload(decode=True)
    file = StringIO.StringIO(data) 
    tar = tarfile.open(None, 'r', fileobj=file)
    sanitized_html = tar.extractfile(index).read()
    assert 'script' not in sanitized_html
    assert tar.extractfile(style).read() == style_data
    assert tar.extractfile(style_screen).read() == style_screen_data
    assert tar.extractfile(style_all).read() == style_all_data
    assert tar.extractfile(style_print).read() == style_print_data
    assert tar.extractfile(image1).read() == image1_data
    assert tar.extractfile(image2).read() == image2_data

def test_multipart_builder(tmpdir):
    create_page(tmpdir)
    builder = response.multipart_builder()
    msg = builder.build(str(tmpdir), index)
    assert msg.is_multipart()
    assert msg['content-type'] == 'multipart/related'
    iterator = msg.walk()
    iterator.next() # skip the multipart
    text = iterator.next()
    modified_data = text.get_payload(decode=True)
    # Styleshets for screen and all should be replace with inline style
    inline_style = '<style type="text/css">' + style_data + '</style>'
    assert inline_style in modified_data
    inline_style_screen = '<style type="text/css">' + style_screen_data + '</style>'
    assert inline_style_screen in modified_data
    inline_style_all = '<style type="text/css">' + style_all_data + '</style>'
    assert inline_style_all in modified_data
    # Other stylesheets should not be inlined
    inline_style_print = '<style type="text/css">' + style_print_data + '</style>'
    assert inline_style_print not in modified_data
    # All link rel="stylesheet" should be removed
    assert '<link rel="stylesheet"' not in modified_data
    # The missing stylesheet should be removed
    # Both references to image1.jpg are replaces with reference to cid:1
    assert '<img src="image1.jpg"' not in modified_data
    assert modified_data.count('<img src="cid:1"') == 2
    # Reference to image2.png is replaced with cid:2
    assert '<img src="/images/image2.png"' not in modified_data
    assert '<img src="cid:2"' in modified_data
    # Missing image is removed from the page
    assert '<img src="missing.gif"' not in modified_data
    assert 'script' not in modified_data
    image1 = iterator.next()
    assert image1['content-type'] == 'image/jpeg'
    assert image1['content-id'] == '<1>'
    assert image1.get_payload(decode=True) == image1_data
    image2 = iterator.next()
    assert image2['content-type'] == 'image/png'
    assert image2['content-id'] == '<2>'
    assert image2.get_payload(decode=True) == image2_data

def test_embded_stylehseet_in_xhtml(tmpdir):
    create_page(tmpdir, type='xhtml')
    builder = response.multipart_builder()
    msg = builder.build(str(tmpdir), index)
    iterator = msg.walk()
    iterator.next() # skip the multipart
    text = iterator.next()
    modified_data = text.get_payload(decode=True)
    escaped_style_data = '''
p :before   { content: "&lt; " }
p :after    { content: " &amp;" }
'''
    inline_style = '<style type="text/css">' + escaped_style_data + '</style>'
    assert inline_style in modified_data
    assert '<link rel="stylesheet"' not in modified_data

